@@ -28,10 +28,6 @@ struct sockopt {
 
 static int filan_streams_analyze(int fd, FILE *outfile);
 
-/* dirty workaround so we dont get an error on AIX when being linked with
-   libwrap */
-int allow_severity, deny_severity;
-
 /* global variables for configuring filan */
 bool filan_followsymlinks;
 bool filan_rawoutput;
@@ -144,6 +140,7 @@ int filan_fd(int fd, FILE *outfile) {
 	    |POLLMSG
 #endif
 	    ;
+#if HAVE_POLL
 	 if (Poll(&ufds, 1, 0) < 0) {
 	    Warn4("poll({%d, %hd, %hd}, 1, 0): %s",
 		   ufds.fd, ufds.events, ufds.revents, strerror(errno));
@@ -200,6 +197,7 @@ int filan_fd(int fd, FILE *outfile) {
 	    }
 #endif /* _WITH_SOCKET && defined(MSG_DONTWAIT) */
 	 }	 
+#endif /* HAVE_POLL */
       }
    }
    fputc('\n', outfile);
@@ -529,20 +527,18 @@ int sockan(int fd, FILE *outfile) {
 #define FILAN_NAMELEN 256
    socklen_t optlen;
    int result /*0, i*/;
-   static const char *socktypes[] = {
-      "undef", "STREAM", "DGRAM", "RAW", "RDM",
-      "SEQPACKET", "undef", "undef", "undef", "undef", 
-      "PACKET", "undef" } ;
    char nambuff[FILAN_NAMELEN];
    /* in Linux these optcodes are 'enum', but on AIX they are bits! */
    static const struct sockopt sockopts[] = {
       {SO_DEBUG, "DEBUG"},
       {SO_REUSEADDR, "REUSEADDR"},
-      {SO_TYPE, "TYPE"},
-      {SO_ERROR, "ERROR"},
-#ifdef SO_PROTOTYPE
+#ifdef SO_PROTOCOL
+      {SO_PROTOCOL, "PROTOCOL"},
+#elif defined(SO_PROTOTYPE)
       {SO_PROTOTYPE, "PROTOTYPE"},
 #endif
+      {SO_TYPE, "TYPE"},
+      {SO_ERROR, "ERROR"},
       {SO_DONTROUTE, "DONTROUTE"},
       {SO_BROADCAST, "BROADCAST"},
       {SO_SNDBUF, "SNDBUF"},
@@ -617,8 +613,12 @@ int sockan(int fd, FILE *outfile) {
       Debug4("getsockopt(%d, SOL_SOCKET, SO_TYPE, %p, {"F_socklen"}): %s",
 	     fd, optval.c, optlen, strerror(errno));
    } else {
+#     define TYPENAMEMAX 16
+      char typename[TYPENAMEMAX];
+      sockettype(*optval.i, typename, sizeof(typename));
+      
       Debug3("fd %d: socket of type %d (\"%s\")", fd, *optval.i,
-	  socktypes[*optval.i]);
+	  typename);
    }
 
    optname = sockopts; while (optname->so) {
@@ -745,6 +745,9 @@ int ipan(int fd, FILE *outfile) {
 #ifdef IP_RECVTOS
       {IP_RECVTOS,      "IP_RECVTOS"},
 #endif
+#ifdef IP_TRANSPARENT
+      {IP_TRANSPARENT,  "IP_TRANSPARENT"},
+#endif
 #ifdef IP_MTU
       {IP_MTU,          "IP_MTU"},
 #endif
@@ -759,22 +762,29 @@ int ipan(int fd, FILE *outfile) {
 #endif
       {0, NULL} } ;
    const struct sockopt *optname;
-   int opttype;
-   socklen_t optlen = sizeof(opttype);
+   int optproto;
+   socklen_t optlen = sizeof(optproto);
    
    optname = ipopts; while (optname->so) {
       sockoptan(fd, optname, SOL_IP, outfile);
       ++optname;
    }
-   /* want to pass the fd to the next layer protocol. dont know how to get the
-      protocol number from the fd? use TYPE to identify TCP. */
-   if (Getsockopt(fd, SOL_SOCKET, SO_TYPE, &opttype, &optlen) >= 0) {
-      switch (opttype) {
+   /* want to pass the fd to the next layer protocol. */
+#if defined(SO_PROTOCOL) || defined(SO_PROTOTYPE)
+   if (Getsockopt(fd, SOL_SOCKET,
+#ifdef SO_PROTOCOL
+		  SO_PROTOCOL,
+#elif defined(SO_PROTOTYPE)
+		  SO_PROTOTYPE,
+#endif
+		  &optproto, &optlen) >= 0) {
+      switch (optproto) {
 #if WITH_TCP
-      case SOCK_STREAM: tcpan(fd, outfile); break;
+      case IPPROTO_TCP: tcpan(fd, outfile); break;
 #endif
       }
    }
+#endif /* defined(SO_PROTOCOL) || defined(SO_PROTOTYPE) */
    return 0;
 }
 #endif /* WITH_IP */
