@@ -1,5 +1,5 @@
 /* source: xio-ip.c */
-/* Copyright Gerhard Rieger */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* this file contains the source for IP related functions */
@@ -14,6 +14,7 @@
 #include "xio-socket.h"
 #include "xio-ip.h"
 #include "xio-ip6.h"
+#include "nestlex.h"
 
 
 #if WITH_IP4 || WITH_IP6
@@ -50,6 +51,9 @@ const struct optdesc opt_ip_mtu_discover={"ip-mtu-discover","mtudiscover",OPT_IP
 #ifdef IP_MTU
 const struct optdesc opt_ip_mtu     = { "ip-mtu",     "mtu",       OPT_IP_MTU,     GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_INT, OFUNC_SOCKOPT, SOL_IP, IP_MTU };
 #endif
+#ifdef IP_TRANSPARENT
+const struct optdesc opt_ip_transparent = {"ip-transparent", "transparent", OPT_IP_TRANSPARENT, GROUP_SOCK_IP, PH_PREBIND, TYPE_BOOL, OFUNC_SOCKOPT, SOL_IP, IP_TRANSPARENT};
+#endif
 #ifdef IP_FREEBIND
 const struct optdesc opt_ip_freebind= { "ip-freebind","freebind",  OPT_IP_FREEBIND,GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_INT, OFUNC_SOCKOPT, SOL_IP, IP_FREEBIND };
 #endif
@@ -67,6 +71,9 @@ const struct optdesc opt_ip_pktoptions = { "ip-pktoptions", "pktopts", OPT_IP_PK
 #ifdef IP_ADD_MEMBERSHIP
 const struct optdesc opt_ip_add_membership = { "ip-add-membership", "membership",OPT_IP_ADD_MEMBERSHIP, GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_IP_MREQN, OFUNC_SOCKOPT, SOL_IP, IP_ADD_MEMBERSHIP };
 #endif
+#ifdef IP_ADD_SOURCE_MEMBERSHIP
+const struct optdesc opt_ip_add_source_membership = { "ip-add-source-membership", "source-membership",OPT_IP_ADD_SOURCE_MEMBERSHIP, GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_IP_MREQ_SOURCE, OFUNC_SOCKOPT, SOL_IP, IP_ADD_SOURCE_MEMBERSHIP };
+#endif
 #ifdef IP_RECVDSTADDR
 const struct optdesc opt_ip_recvdstaddr = { "ip-recvdstaddr", "recvdstaddr",OPT_IP_RECVDSTADDR, GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_INT, OFUNC_SOCKOPT, SOL_IP, IP_RECVDSTADDR };
 #endif
@@ -74,11 +81,19 @@ const struct optdesc opt_ip_recvdstaddr = { "ip-recvdstaddr", "recvdstaddr",OPT_
 const struct optdesc opt_ip_recvif = { "ip-recvif", "recvdstaddrif",OPT_IP_RECVIF, GROUP_SOCK_IP, PH_PASTSOCKET, TYPE_INT, OFUNC_SOCKOPT, SOL_IP, IP_RECVIF };
 #endif
 
+#if WITH_RES_DEPRECATED
+#  define WITH_RES_AAONLY 1
+#  define WITH_RES_PRIMARY 1
+#endif /* WITH_RES_DEPRECATED */
 #if HAVE_RESOLV_H
 const struct optdesc opt_res_debug    = { "res-debug",    NULL,       OPT_RES_DEBUG,    GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_DEBUG };
+#if WITH_RES_AAONLY
 const struct optdesc opt_res_aaonly   = { "res-aaonly",   "aaonly",   OPT_RES_AAONLY,   GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_AAONLY };
+#endif
 const struct optdesc opt_res_usevc    = { "res-usevc",    "usevc",    OPT_RES_USEVC,    GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_USEVC };
+#if WITH_RES_PRIMARY
 const struct optdesc opt_res_primary  = { "res-primary",  "primary",  OPT_RES_PRIMARY,  GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_PRIMARY };
+#endif
 const struct optdesc opt_res_igntc    = { "res-igntc",    "igntc",    OPT_RES_IGNTC,    GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_IGNTC };
 const struct optdesc opt_res_recurse  = { "res-recurse",  "recurse",  OPT_RES_RECURSE,  GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_RECURSE };
 const struct optdesc opt_res_defnames = { "res-defnames", "defnames", OPT_RES_DEFNAMES, GROUP_SOCK_IP, PH_INIT, TYPE_BOOL, OFUNC_OFFSET_MASKS, XIO_OFFSETOF(para.socket.ip.res_opts), XIO_SIZEOF(para.socket.ip.res_opts), RES_DEFNAMES };
@@ -113,7 +128,7 @@ unsigned long res_opts() {
    hostname.domain (fq hostname resolving to IPv4 or IPv6 address)
  service: the port specification; may be numeric or symbolic
  family: PF_INET, PF_INET6, or PF_UNSPEC permitting both
- socktype: SOCK_STREAM, SOCK_DGRAM
+ socktype: SOCK_STREAM, SOCK_DGRAM, ...
  protocol: IPPROTO_UDP, IPPROTO_TCP
  sau: an uninitialized storage for the resulting socket address
  returns: STAT_OK, STAT_RETRYLATER
@@ -152,6 +167,17 @@ int xiogetaddrinfo(const char *node, const char *service,
    if (service && service[0]=='\0') {
       Error("empty port/service");
    }
+
+#ifdef WITH_VSOCK
+   if (family == AF_VSOCK) {
+      error_num = sockaddr_vm_parse(&sau->vm, node, service);
+      if (error_num < 0)
+         return STAT_NORETRY;
+
+      return STAT_OK;
+   }
+#endif /* WITH_VSOCK */
+
    /* if service is numeric we don't want to have a lookup (might take long
       with NIS), so we handle this specially */
    if (service && isdigit(service[0]&0xff)) {
@@ -167,30 +193,8 @@ int xiogetaddrinfo(const char *node, const char *service,
    /* the resolver functions might handle numeric forms of node names by
       reverse lookup, that's not what we want.
       So we detect these and handle them specially */
-   if (node && isdigit(node[0]&0xff)) {
-#if HAVE_GETADDRINFO
-      hints.ai_flags |= AI_NUMERICHOST;
-#endif /* HAVE_GETADDRINFO */
-      if (family == PF_UNSPEC) {
-	 family = PF_INET;
-#if HAVE_GETADDRINFO
-      } else if (family == PF_INET6) {
-	 /* map "explicitely" into IPv6 address space; getipnodebyname() does
-	    this with AI_V4MAPPED, but not getaddrinfo() */
-	 if ((numnode = Malloc(strlen(node)+7+1)) == NULL) {
-#if HAVE_RESOLV_H
-	    if (res_opts0 | res_opts1) {
-	       _res.options = (_res.options & (~res_opts0&~res_opts1) |
-			       save_res_opts& ( res_opts0| res_opts1));
-	    }
-#endif
-	    return STAT_NORETRY;
-	 }
-	 sprintf(numnode, "::ffff:%s", node);
-	 node = numnode;
-	 hints.ai_flags |= AI_NUMERICHOST;
-#endif /* HAVE_GETADDRINFO */
-      }
+   if (0) { 	/* for canonical reasons */
+      ;
 #if WITH_IP6
    } else if (node && node[0] == '[' && node[(nodelen=strlen(node))-1]==']') {
       if ((numnode = Malloc(nodelen-1)) == NULL) {
@@ -216,18 +220,6 @@ int xiogetaddrinfo(const char *node, const char *service,
    if (node != NULL || service != NULL) {
       struct addrinfo *record;
 
-      if (socktype != SOCK_STREAM && socktype != SOCK_DGRAM) {
-	 /* actual socket type value is not supported - fallback to a good one */
-	 socktype = SOCK_DGRAM;
-      }
-      if (protocol != IPPROTO_TCP && protocol != IPPROTO_UDP) {
-	 /* actual protocol value is not supported - fallback to a good one */
-	 if (socktype == SOCK_DGRAM) {
-	    protocol = IPPROTO_UDP;
-	 } else {
-	    protocol = IPPROTO_TCP;
-	 }
-      }
       hints.ai_flags |= AI_PASSIVE;
       hints.ai_family = family;
       hints.ai_socktype = socktype;
@@ -237,9 +229,22 @@ int xiogetaddrinfo(const char *node, const char *service,
       hints.ai_canonname = NULL;
       hints.ai_next = NULL;
 
-      if ((error_num = Getaddrinfo(node, service, &hints, &res)) != 0) {
+      do {
+	error_num = Getaddrinfo(node, service, &hints, &res);
+	if (error_num == 0)  break;
+	if (error_num == EAI_SOCKTYPE && socktype != 0) {
+	   /* there are systems where kernel goes SCTP but not getaddrinfo() */
+	   hints.ai_socktype = 0;
+	   continue;
+	}
+	if (error_num == EAI_SERVICE && protocol != 0) {
+	   hints.ai_protocol = 0;
+	   continue;
+	}
+	if (error_num != 0) {
 	 Error7("getaddrinfo(\"%s\", \"%s\", {%d,%d,%d,%d}, {}): %s",
-		node, service, hints.ai_flags, hints.ai_family,
+		node?node:"NULL", service?service:"NULL",
+		hints.ai_flags, hints.ai_family,
 		hints.ai_socktype, hints.ai_protocol,
 		(error_num == EAI_SYSTEM)?
 		strerror(errno):gai_strerror(error_num));
@@ -253,7 +258,8 @@ int xiogetaddrinfo(const char *node, const char *service,
 	 }
 #endif
 	 return STAT_RETRYLATER;
-      }
+	}
+      } while (1);
       service = NULL;	/* do not resolve later again */
 
       record = res;
@@ -371,7 +377,7 @@ int xiogetaddrinfo(const char *node, const char *service,
       /* this is not a typical IP6 resolver function - but Linux
 	 "man gethostbyname" says that the only supported address type with
 	 this function is AF_INET _at present_, so maybe this fallback will
-	 be useful somewhere sometimesin a future even for IP6 */
+	 be useful somewhere sometimes in a future even for IP6 */
       if (family == PF_UNSPEC) {
 #if WITH_IP4 && WITH_IP6
 	 family = xioopts.default_ip=='6'?PF_INET6:PF_INET;
@@ -464,7 +470,8 @@ int xiolog_ancillary_ip(struct cmsghdr *cmsg, int *num,
 			char *nambuff, int namlen,
 			char *envbuff, int envlen,
 			char *valbuff, int vallen) {
-   const char *cmsgtype, *cmsgname = NULL, *cmsgenvn = NULL, *cmsgfmt = NULL;
+   int cmsgctr = 0;
+   const char *cmsgtype, *cmsgname = NULL, *cmsgenvn = NULL;
    size_t msglen;
    char scratch1[16];	/* can hold an IPv4 address in ASCII */
 #if WITH_IP4 && defined(IP_PKTINFO) && HAVE_STRUCT_IN_PKTINFO
@@ -554,17 +561,17 @@ int xiolog_ancillary_ip(struct cmsghdr *cmsg, int *num,
 #ifdef IP_RECVOPTS
    case IP_RECVOPTS:
 #endif
-      cmsgtype = "IP_OPTIONS"; cmsgname = "options"; cmsgfmt = NULL; break;
+      cmsgtype = "IP_OPTIONS"; cmsgname = "options"; cmsgctr = -1; break;
    case IP_TOS:
-      cmsgtype = "IP_TOS";     cmsgname = "tos"; cmsgfmt = "%u"; break;
+      cmsgtype = "IP_TOS";     cmsgname = "tos"; cmsgctr = msglen; break;
    case IP_TTL: /* Linux */
 #ifdef IP_RECVTTL
    case IP_RECVTTL: /* FreeBSD */
 #endif
-      cmsgtype = "IP_TTL";     cmsgname = "ttl"; cmsgfmt = "%u"; break;
+      cmsgtype = "IP_TTL";     cmsgname = "ttl"; cmsgctr = msglen; break;
    }
    /* when we come here we provide a single parameter
-      with type in cmsgtype, name in cmsgname, printf format in cmsgfmt */
+      with type in cmsgtype, name in cmsgname, value length in msglen */
    *num = 1;
    if (strlen(cmsgtype) >= typlen)  rc = STAT_WARNING;
    typbuff[0] = '\0'; strncat(typbuff, cmsgtype, typlen-1);
@@ -576,13 +583,141 @@ int xiolog_ancillary_ip(struct cmsghdr *cmsg, int *num,
    } else {
       envbuff[0] = '\0';
    }
-   if (cmsgfmt != NULL) {
-      snprintf(valbuff, vallen, cmsgfmt, *(unsigned char *)CMSG_DATA(cmsg));
-   } else {
-      xiodump(CMSG_DATA(cmsg), msglen, valbuff, vallen, 0);
+   switch (cmsgctr) {
+   case sizeof(char):
+      snprintf(valbuff, vallen, "%u", *(unsigned char *)CMSG_DATA(cmsg)); break;
+   case sizeof(int):
+      snprintf(valbuff, vallen, "%u",    (*(unsigned int *)CMSG_DATA(cmsg))); break;
+   case 0:
+      xiodump(CMSG_DATA(cmsg), msglen, valbuff, vallen, 0); break;
+   default: break;
    }
    return rc;
 }
 #endif /* defined(HAVE_STRUCT_CMSGHDR) && defined(CMSG_DATA) */
 
+
+#if HAVE_STRUCT_IP_MREQ_SOURCE
+int xiotype_ip_add_source_membership(char *token, const struct optname *ent, struct opt *opt) {
+   /* we do not resolve the addresses here because we do not yet know
+      if we are coping with an IPv4 or IPv6 socat address */
+   const char *ends[] = { ":", NULL };
+   const char *nests[] = { "[","]", NULL };
+   char buff[512], *buffp=buff; size_t bufspc = sizeof(buff)-1;
+   char *tokp = token;
+   int parsres;
+
+   /* parse first IP address, expect ':' */
+   parsres =
+      nestlex((const char **)&tokp, &buffp, &bufspc,
+	      ends, NULL, NULL, nests,
+	      true, false, false);
+   if (parsres < 0) {
+      Error1("option too long:  \"%s\"", token);
+      return -1;
+   } else if (parsres > 0) {
+      Error1("syntax error in \"%s\"", token);
+      return -1;
+   }
+   if (*tokp != ':') {
+      Error1("syntax in option %s: missing ':'", token);
+   }
+   *buffp++ = '\0';
+   opt->value.u_ip_mreq_source.mcaddr = strdup(buff); /*!!! NULL */
+
+   ++tokp;
+   /* parse second IP address, expect ':' or '\0'' */
+   buffp = buff;
+   /*! result= */
+   parsres =
+      nestlex((const char **)&tokp, &buffp, &bufspc,
+	      ends, NULL, NULL, nests,
+	      true, false, false);
+   if (parsres < 0) {
+      Error1("option too long:  \"%s\"", token);
+      return -1;
+   } else if (parsres > 0) {
+      Error1("syntax error in \"%s\"", token);
+      return -1;
+   }
+   if (*tokp != ':') {
+      Error1("syntax in option %s: missing ':'", token);
+   }
+   *buffp++ = '\0';
+   opt->value.u_ip_mreq_source.ifaddr = strdup(buff); /*!!! NULL */
+
+   ++tokp;
+   /* parse third IP address, expect ':' or '\0'' */
+   buffp = buff;
+   /*! result= */
+   parsres =
+      nestlex((const char **)&tokp, &buffp, &bufspc,
+	      ends, NULL, NULL, nests,
+	      true, false, false);
+   if (parsres < 0) {
+      Error1("option too long:  \"%s\"", token);
+      return -1;
+   } else if (parsres > 0) {
+      Error1("syntax error in \"%s\"", token);
+      return -1;
+   }
+   if (*tokp) {
+      Error1("syntax in option %s: trailing cruft", token);
+   }
+   *buffp++ = '\0';
+   opt->value.u_ip_mreq_source.srcaddr = strdup(buff); /*!!! NULL */
+
+   Info4("setting option \"%s\" to {0x%08x,0x%08x,0x%08x}",
+	 ent->desc->defname,
+	 ntohl(*(unsigned int *)opt->value.u_ip_mreq_source.mcaddr),
+	 ntohl(*(unsigned int *)opt->value.u_ip_mreq_source.ifaddr),
+	 ntohl(*(unsigned int *)opt->value.u_ip_mreq_source.srcaddr));
+   return 0;
+}
+
+int xioapply_ip_add_source_membership(struct single *xfd, struct opt *opt) {
+   struct ip_mreq_source ip4_mreq_src = {{0}};
+   /* IPv6 not supported - seems to have different handling */
+   union sockaddr_union sockaddr1;
+   socklen_t socklen1 = sizeof(sockaddr1.ip4);
+   union sockaddr_union sockaddr2;
+   socklen_t socklen2 = sizeof(sockaddr2.ip4);
+   union sockaddr_union sockaddr3;
+   socklen_t socklen3 = sizeof(sockaddr3.ip4);
+
+   /* first parameter is always multicast address */
+   /*! result */
+   xiogetaddrinfo(opt->value.u_ip_mreq_source.mcaddr, NULL,
+		  xfd->para.socket.la.soa.sa_family,
+		  SOCK_DGRAM, IPPROTO_IP,
+		  &sockaddr1, &socklen1, 0, 0);
+   ip4_mreq_src.imr_multiaddr = sockaddr1.ip4.sin_addr;
+   /* second parameter is interface address */
+   xiogetaddrinfo(opt->value.u_ip_mreq_source.ifaddr, NULL,
+		  xfd->para.socket.la.soa.sa_family,
+		  SOCK_DGRAM, IPPROTO_IP,
+		  &sockaddr2, &socklen2, 0, 0);
+   ip4_mreq_src.imr_interface = sockaddr2.ip4.sin_addr;
+   /* third parameter is source address */
+   xiogetaddrinfo(opt->value.u_ip_mreq_source.srcaddr, NULL,
+		  xfd->para.socket.la.soa.sa_family,
+		  SOCK_DGRAM, IPPROTO_IP,
+		  &sockaddr3, &socklen3, 0, 0);
+   ip4_mreq_src.imr_sourceaddr = sockaddr3.ip4.sin_addr;
+   if (Setsockopt(xfd->fd, opt->desc->major, opt->desc->minor,
+		  &ip4_mreq_src, sizeof(ip4_mreq_src)) < 0) {
+      Error8("setsockopt(%d, %d, %d, {0x%08x,0x%08x,0x%08x}, "F_Zu"): %s",
+	     xfd->fd, opt->desc->major, opt->desc->minor,
+	     htonl((uint32_t)ip4_mreq_src.imr_multiaddr.s_addr),
+	     ip4_mreq_src.imr_interface.s_addr,
+	     ip4_mreq_src.imr_sourceaddr.s_addr,
+	     sizeof(struct ip_mreq_source),
+	     strerror(errno));
+      opt->desc = ODESC_ERROR;
+      return -1;
+   }
+   return 0;
+}
+#endif /* HAVE_STRUCT_IP_MREQ_SOURCE */
+
 #endif /* _WITH_IP4 || _WITH_IP6 */
