@@ -1,5 +1,5 @@
-/* $Id: xio-proxy.h,v 1.6 2006/05/31 19:17:57 gerhard Exp $ */
-/* Copyright Gerhard Rieger 2002-2006 */
+/* source: xio-proxy.h */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 #ifndef __xio_proxy_h_included
@@ -10,6 +10,7 @@ struct proxyvars {
    bool ignorecr;
    bool doresolve;
    char *authstring;
+   char *authfile;
    char *targetaddr;	/* name/address of host, in malloced string */
    uint16_t targetport;
 } ;
@@ -18,6 +19,7 @@ extern const struct optdesc opt_proxyport;
 extern const struct optdesc opt_ignorecr;
 extern const struct optdesc opt_proxy_resolve;
 extern const struct optdesc opt_proxy_authorization;
+extern const struct optdesc opt_proxy_authorization_file;
 
 extern const struct addrdesc addr_proxy_connect;
 
