@@ -173,7 +173,7 @@ const struct optdesc opt_so_dontlinger = {"so-dontlinger", "dontlinger",  OPT_SO
 #endif
 /* the SO_PROTOTYPE is OS defined on Solaris, HP-UX; we lend this for a more
    general purpose */
-const struct optdesc opt_so_prototype  = {"so-prototype",  "prototype",   OPT_SO_PROTOTYPE,   GROUP_SOCKET,PH_SOCKET,    TYPE_INT,OFUNC_SPEC,   SOL_SOCKET,SO_PROTOTYPE };
+const struct optdesc opt_so_prototype  = {"so-protocol",  "protocol",   OPT_SO_PROTOTYPE,   GROUP_SOCKET,PH_SOCKET,    TYPE_INT,OFUNC_SPEC,   SOL_SOCKET,SO_PROTOCOL };
 #ifdef FIOSETOWN
 const struct optdesc opt_fiosetown   = { "fiosetown", NULL, OPT_FIOSETOWN,   GROUP_SOCKET, PH_PASTSOCKET, TYPE_INT,  OFUNC_IOCTL,  FIOSETOWN };
 #endif
@@ -186,9 +186,11 @@ const struct optdesc opt_protocol_family = { "protocol-family", "pf", OPT_PROTOC
 const struct optdesc opt_protocol        = { "protocol",        NULL, OPT_PROTOCOL,        GROUP_SOCKET, PH_PRESOCKET,  TYPE_STRING,  OFUNC_SPEC };
 
 /* generic setsockopt() options */
-const struct optdesc opt_setsockopt_int    = { "setsockopt-int",    "sockopt-int",    OPT_SETSOCKOPT_INT,        GROUP_SOCKET,PH_PASTSOCKET,TYPE_INT_INT_INT,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
-const struct optdesc opt_setsockopt_bin    = { "setsockopt-bin",    "sockopt-bin",    OPT_SETSOCKOPT_BIN,        GROUP_SOCKET,PH_PASTSOCKET,TYPE_INT_INT_BIN,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
-const struct optdesc opt_setsockopt_string = { "setsockopt-string", "sockopt-string", OPT_SETSOCKOPT_STRING,     GROUP_SOCKET,PH_PASTSOCKET,TYPE_INT_INT_STRING,  OFUNC_SOCKOPT_GENERIC, 0, 0 };
+const struct optdesc opt_setsockopt        = { "setsockopt",        "sockopt",        OPT_SETSOCKOPT_BIN,        GROUP_SOCKET,PH_CONNECTED, TYPE_INT_INT_BIN,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
+const struct optdesc opt_setsockopt_int    = { "setsockopt-int",    "sockopt-int",    OPT_SETSOCKOPT_INT,        GROUP_SOCKET,PH_CONNECTED, TYPE_INT_INT_INT,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
+const struct optdesc opt_setsockopt_bin    = { "setsockopt-bin",    "sockopt-bin",    OPT_SETSOCKOPT_BIN,        GROUP_SOCKET,PH_CONNECTED, TYPE_INT_INT_BIN,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
+const struct optdesc opt_setsockopt_string = { "setsockopt-string", "sockopt-string", OPT_SETSOCKOPT_STRING,     GROUP_SOCKET,PH_CONNECTED, TYPE_INT_INT_STRING,  OFUNC_SOCKOPT_GENERIC, 0, 0 };
+const struct optdesc opt_setsockopt_listen = { "setsockopt-listen", "sockopt-listen", OPT_SETSOCKOPT_LISTEN,     GROUP_SOCKET,PH_PREBIND,   TYPE_INT_INT_BIN,     OFUNC_SOCKOPT_GENERIC, 0, 0 };
 
 const struct optdesc opt_null_eof = { "null-eof", NULL, OPT_NULL_EOF, GROUP_SOCKET, PH_INIT, TYPE_BOOL, OFUNC_OFFSET, XIO_OFFSETOF(para.socket.null_eof) };
 
@@ -239,7 +241,7 @@ int xioopen_socket_connect(int argc, const char *argv[], struct opt *opts,
 
    themsize = 0;
    if ((result =
-	dalan(address, (char *)&them.soa.sa_data, &themsize, sizeof(them)))
+	dalan(address, (uint8_t *)&them.soa.sa_data, &themsize, sizeof(them), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", address);
    } else if (result > 0) {
@@ -264,7 +266,7 @@ int xioopen_socket_connect(int argc, const char *argv[], struct opt *opts,
 
    if ((result =
 	xioopen_connect(xfd,
-			needbind?(struct sockaddr *)&us:NULL, uslen,
+			needbind?&us:NULL, uslen,
 			(struct sockaddr *)&them, themlen,
 			opts, pf, socktype, proto, false)) != 0) {
       return result;
@@ -316,7 +318,7 @@ int xioopen_socket_listen(int argc, const char *argv[], struct opt *opts,
    socket_init(0, &us);
    ussize = 0;
    if ((result =
-	dalan(usname, (char *)&us.soa.sa_data, &ussize, sizeof(us)))
+	dalan(usname, (uint8_t *)&us.soa.sa_data, &ussize, sizeof(us), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", usname);
    } else if (result > 0) {
@@ -337,7 +339,7 @@ int xioopen_socket_listen(int argc, const char *argv[], struct opt *opts,
 
    if ((result =
 	xioopen_listen(xfd, xioflags,
-		       (struct sockaddr *)&us, uslen,
+		       &us.soa, uslen,
 		       opts, opts0, 0/*instead of pf*/, socktype, proto))
        != STAT_OK)
       return result;
@@ -407,8 +409,8 @@ int _xioopen_socket_sendto(const char *pfname, const char *type,
    xfd->peersa.soa.sa_family = pf;
    themsize = 0;
    if ((result =
-	dalan(address, (char *)&xfd->peersa.soa.sa_data, &themsize,
-	      sizeof(xfd->peersa)))
+	dalan(address, (uint8_t *)&xfd->peersa.soa.sa_data, &themsize,
+	      sizeof(xfd->peersa), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", address);
    } else if (result > 0) {
@@ -438,7 +440,7 @@ int _xioopen_socket_sendto(const char *pfname, const char *type,
    if (retropt_string(opts, OPT_BIND, &bindstring) == 0) {
       ussize = 0;
       if ((result =
-	   dalan(bindstring, (char *)&us.soa.sa_data, &ussize, sizeof(us)))
+	   dalan(bindstring, (uint8_t *)&us.soa.sa_data, &ussize, sizeof(us), 'i'))
 	  < 0) {
 	 Error1("data too long: \"%s\"", bindstring);
       } else if (result > 0) {
@@ -504,7 +506,7 @@ int xioopen_socket_recvfrom(int argc, const char *argv[], struct opt *opts,
 
    ussize = 0;
    if ((result =
-	dalan(address, (char *)&us->soa.sa_data, &ussize, sizeof(*us)))
+	dalan(address, (uint8_t *)&us->soa.sa_data, &ussize, sizeof(*us), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", address);
    } else if (result > 0) {
@@ -581,7 +583,7 @@ int xioopen_socket_recv(int argc, const char *argv[], struct opt *opts,
 
    ussize = 0;
    if ((result =
-	dalan(address, (char *)&us.soa.sa_data, &ussize, sizeof(us)))
+	dalan(address, (uint8_t *)&us.soa.sa_data, &ussize, sizeof(us), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", address);
    } else if (result > 0) {
@@ -649,8 +651,8 @@ int xioopen_socket_datagram(int argc, const char *argv[], struct opt *opts,
    xfd->peersa.soa.sa_family = pf;
    themsize = 0;
    if ((result =
-	dalan(address, (char *)&xfd->peersa.soa.sa_data, &themsize,
-	      sizeof(xfd->peersa)))
+	dalan(address, (uint8_t *)&xfd->peersa.soa.sa_data, &themsize,
+	      sizeof(xfd->peersa), 'i'))
        < 0) {
       Error1("data too long: \"%s\"", address);
    } else if (result > 0) {
@@ -702,7 +704,7 @@ int xioopen_socket_datagram(int argc, const char *argv[], struct opt *opts,
    OPT_SO_TYPE, OPT_SO_PROTOTYPE, OPT_USER, OPT_GROUP, OPT_CLOEXEC
    returns 0 on success.
 */
-int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
+int _xioopen_connect(struct single *xfd, union sockaddr_union *us, size_t uslen,
 		     struct sockaddr *them, size_t themlen,
 		     struct opt *opts, int pf, int socktype, int protocol,
 		     bool alt, int level) {
@@ -713,6 +715,12 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
    int _errno;
    int result;
 
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_EARLY);
+   }
+#endif
+
    if ((xfd->fd = xiosocket(opts, pf, socktype, protocol, level)) < 0) {
       return STAT_RETRYLATER;	    
    }
@@ -723,6 +731,11 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
 
    applyopts_cloexec(xfd->fd, opts);
 
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_PREOPEN);
+   }
+#endif
    applyopts(xfd->fd, opts, PH_PREBIND);
    applyopts(xfd->fd, opts, PH_BIND);
 #if WITH_TCP || WITH_UDP
@@ -733,7 +746,7 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
 
       /* prepare sockaddr for bind probing */
       if (us) {
-	 sinp = (union sockaddr_union *)us;
+	 sinp = us;
       } else {
 	 if (them->sa_family == AF_INET) {
 	    socket_in_init(&sin.ip4);
@@ -779,7 +792,7 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
       i = N = XIO_IPPORT_LOWER + dv.rem;
       do {	/* loop over lowport bind() attempts */
 	 *port = htons(i);
-	 if (Bind(xfd->fd, (struct sockaddr *)sinp, sizeof(*sinp)) < 0) {
+	 if (Bind(xfd->fd, &sinp->soa, sizeof(*sinp)) < 0) {
 	    Msg4(errno==EADDRINUSE?E_INFO:level,
 		 "bind(%d, {%s}, "F_Zd"): %s", xfd->fd,
 		 sockaddr_info(&sinp->soa, sizeof(*sinp), infobuff, sizeof(infobuff)),
@@ -803,14 +816,24 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
 #endif /* WITH_TCP || WITH_UDP */
 
    if (us) {
-      if (Bind(xfd->fd, us, uslen) < 0) {
+#if WITH_UNIX
+      if (pf == PF_UNIX && us != NULL) {
+	 applyopts_named(us->un.sun_path, opts, PH_PREOPEN);
+      }
+#endif
+      if (Bind(xfd->fd, &us->soa, uslen) < 0) {
 	 Msg4(level, "bind(%d, {%s}, "F_Zd"): %s",
-	      xfd->fd, sockaddr_info(us, uslen, infobuff, sizeof(infobuff)),
+	      xfd->fd, sockaddr_info(&us->soa, uslen, infobuff, sizeof(infobuff)),
 	      uslen, strerror(errno));
 	 Close(xfd->fd);
 	 return STAT_RETRYLATER;
       }
    }
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_PASTOPEN);
+   }
+#endif
 
    applyopts(xfd->fd, opts, PH_PASTBIND);
 
@@ -822,7 +845,7 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
       Fcntl_l(xfd->fd, F_SETFL, fcntl_flags|O_NONBLOCK);
    }
 
-   result = Connect(xfd->fd, (struct sockaddr *)them, themlen);
+   result = Connect(xfd->fd, them, themlen);
    _errno = errno;
    la.soa.sa_family = them->sa_family;  lalen = sizeof(la);
    if (Getsockname(xfd->fd, &la.soa, &lalen) < 0) {
@@ -909,6 +932,11 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
 
    applyopts_fchown(xfd->fd, opts);	/* OPT_USER, OPT_GROUP */
    applyopts(xfd->fd, opts, PH_CONNECTED);
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_LATE);
+   }
+#endif
    applyopts(xfd->fd, opts, PH_LATE);
 
    return STAT_OK;
@@ -925,7 +953,7 @@ int _xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
    OPT_FORK, OPT_SO_TYPE, OPT_SO_PROTOTYPE, OPT_USER, OPT_GROUP, OPT_CLOEXEC
    returns 0 on success.
 */
-int xioopen_connect(struct single *xfd, struct sockaddr *us, size_t uslen,
+int xioopen_connect(struct single *xfd, union sockaddr_union *us, size_t uslen,
 		    struct sockaddr *them, size_t themlen,
 		    struct opt *opts, int pf, int socktype, int protocol,
 		    bool alt) {
@@ -1033,6 +1061,12 @@ int _xioopen_dgram_sendto(/* them is already in xfd->peersa */
    union sockaddr_union la; socklen_t lalen = sizeof(la);
    char infobuff[256];
 
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_EARLY);
+   }
+#endif
+
    if ((xfd->fd = xiosocket(opts, pf, socktype, ipproto, level)) < 0) {
       return STAT_RETRYLATER;	    
    }
@@ -1044,18 +1078,28 @@ int _xioopen_dgram_sendto(/* them is already in xfd->peersa */
 
    applyopts_cloexec(xfd->fd, opts);
 
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_PREOPEN);
+   }
+#endif
    applyopts(xfd->fd, opts, PH_PREBIND);
    applyopts(xfd->fd, opts, PH_BIND);
 
    if (us) {
-      if (Bind(xfd->fd, (struct sockaddr *)us, uslen) < 0) {
+      if (Bind(xfd->fd, &us->soa, uslen) < 0) {
 	 Msg4(level, "bind(%d, {%s}, "F_socklen"): %s",
-	      xfd->fd, sockaddr_info((struct sockaddr *)us, uslen, infobuff, sizeof(infobuff)),
+	      xfd->fd, sockaddr_info(&us->soa, uslen, infobuff, sizeof(infobuff)),
 	      uslen, strerror(errno));
 	 Close(xfd->fd);
 	 return STAT_RETRYLATER;
       }
    }
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_PASTOPEN);
+   }
+#endif
 
    applyopts(xfd->fd, opts, PH_PASTBIND);
 
@@ -1068,6 +1112,11 @@ int _xioopen_dgram_sendto(/* them is already in xfd->peersa */
 
    applyopts_fchown(xfd->fd, opts);
    applyopts(xfd->fd, opts, PH_CONNECTED);
+#if WITH_UNIX
+   if (pf == PF_UNIX && us != NULL) {
+      applyopts_named(us->un.sun_path, opts, PH_LATE);
+   }
+#endif
    applyopts(xfd->fd, opts, PH_LATE);
 
    /* xfd->dtype = DATA_RECVFROM; *//* no, the caller must set this (ev _SKIPIP) */
@@ -1097,6 +1146,8 @@ static pid_t xio_waitingfor;	/* info from recv loop to signal handler:
 				   that should send us the USR1 signal */
 static bool xio_hashappened;	/* info from signal handler to loop: child
 				   process has read ("consumed") the packet */
+static int xio_childstatus;
+
 /* this is the signal handler for USR1 and CHLD */
 void xiosigaction_hasread(int signum
 #if HAVE_STRUCT_SIGACTION_SA_SIGACTION && defined(SA_SIGINFO)
@@ -1148,6 +1199,7 @@ void xiosigaction_hasread(int signum
 	 }
 	 if (pid == xio_waitingfor) {
 	    xio_hashappened = true;
+            xio_childstatus = WEXITSTATUS(status);
 	    Debug("xiosigaction_hasread() ->");
 	    diag_in_handler = 0;
 	    errno = _errno;
@@ -1221,7 +1273,7 @@ int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
 
    applyopts(xfd->fd, opts, PH_PREBIND);
    applyopts(xfd->fd, opts, PH_BIND);
-   if ((us != NULL) && Bind(xfd->fd, (struct sockaddr *)us, uslen) < 0) {
+   if ((us != NULL) && Bind(xfd->fd, us, uslen) < 0) {
       Msg4(level, "bind(%d, {%s}, "F_socklen"): %s", xfd->fd,
 	   sockaddr_info(us, uslen, infobuff, sizeof(infobuff)), uslen,
 	   strerror(errno));
@@ -1362,7 +1414,7 @@ int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
       palen = msgh.msg_namelen;
 
       Notice1("receiving packet from %s"/*"src"*/,
-	      sockaddr_info((struct sockaddr *)pa, palen, peername, sizeof(peername))/*,
+	      sockaddr_info(&pa->soa, palen, peername, sizeof(peername))/*,
 							     sockaddr_info(&la->soa, sockname, sizeof(sockname))*/);
 
       xiodopacketinfo(&msgh, true, true);
@@ -1374,7 +1426,7 @@ int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
 	 continue;
       }
       Info1("permitting packet from %s",
-	    sockaddr_info((struct sockaddr *)pa, palen,
+	    sockaddr_info(&pa->soa, palen,
 			  infobuff, sizeof(infobuff)));
 
       /* set the env vars describing the local and remote sockets */
@@ -1389,27 +1441,27 @@ int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
       xfd->salen = palen;
 
       if (dofork) {
-	 sigset_t mask_sigchldusr1;
+	 sigset_t oldset, mask_sigchldusr1;
 
 	 /* we must prevent that the current packet triggers another fork;
 	    therefore we wait for a signal from the recent child: USR1
 	    indicates that is has consumed the last packet; CHLD means it has
 	    terminated */
 	 /* block SIGCHLD and SIGUSR1 until parent is ready to react */
-	 sigemptyset(&mask_sigchldusr1);
+	 Sigprocmask(SIG_BLOCK, NULL, &mask_sigchldusr1);
 	 sigaddset(&mask_sigchldusr1, SIGCHLD);
 	 sigaddset(&mask_sigchldusr1, SIGUSR1);
-	 Sigprocmask(SIG_BLOCK, &mask_sigchldusr1, NULL);
+	 Sigprocmask(SIG_SETMASK, &mask_sigchldusr1, &oldset);
 
 	 if ((pid = xio_fork(false, level)) < 0) {
 	    Close(xfd->fd);
-	    Sigprocmask(SIG_UNBLOCK, &mask_sigchldusr1, NULL);
+	    Sigprocmask(SIG_SETMASK, &oldset, NULL);
 	    return STAT_RETRYLATER;
 	 }
 
 	 if (pid == 0) {	/* child */
 	    /* no reason to block SIGCHLD in child process */
-	    Sigprocmask(SIG_UNBLOCK, &mask_sigchldusr1, NULL);
+	    Sigprocmask(SIG_SETMASK, &oldset, NULL);
 	    xfd->ppid = Getppid();	/* send parent a signal when packet has
 					   been consumed */
 
@@ -1431,15 +1483,30 @@ int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
 
 	 /* server: continue loop with listen */
 	 xio_waitingfor = pid;
+
+#if HAVE_PSELECT
+	 {
+	    struct timespec timeout = { LONG_MAX, 0 };
+	    Pselect(0, NULL, NULL, NULL, &timeout, &oldset);
+	    Sigprocmask(SIG_SETMASK, &oldset, NULL);
+	 }
+#else /* ! HAVE_PSELECT */
 	 /* now we are ready to handle signals */
-	 Sigprocmask(SIG_UNBLOCK, &mask_sigchldusr1, NULL);
+	 Sigprocmask(SIG_SETMASK, &oldset, NULL);
 
 	 while (!xio_hashappened) {
-	    Sleep(UINT_MAX);	/* any signal lets us continue */
+	    Sleep(1);	/* any signal speeds up return */
 	 }
+#endif /* ! HAVE_PSELECT */
 	 xio_waitingfor = 0;	/* so this child will not set hashappened again */
 	 xio_hashappened = false;
 
+         if (xio_childstatus != 0) {
+             char buff[512];
+             Recv(xfd->fd, buff, sizeof(buff), 0);
+             xio_childstatus = 0;
+             Info("drop data because of child exit failed");
+         }
 	 Info("continue listening");
       } else {
 	break;
@@ -1473,7 +1540,7 @@ int _xioopen_dgram_recv(struct single *xfd, int xioflags,
 
    applyopts(xfd->fd, opts, PH_PREBIND);
    applyopts(xfd->fd, opts, PH_BIND);
-   if ((us != NULL) && Bind(xfd->fd, (struct sockaddr *)us, uslen) < 0) {
+   if ((us != NULL) && Bind(xfd->fd, us, uslen) < 0) {
       Msg4(level, "bind(%d, {%s}, "F_socklen"): %s", xfd->fd,
 	   sockaddr_info(us, uslen, infobuff, sizeof(infobuff)), uslen,
 	   strerror(errno));
@@ -1718,12 +1785,12 @@ int xiocheckpeer(xiosingle_t *xfd,
       if (xiocheckrange(pa, &xfd->para.socket.range) < 0) {
 	 char infobuff[256];
 	 Warn1("refusing connection from %s due to range option",
-	       sockaddr_info((struct sockaddr *)pa, 0,
+	       sockaddr_info(&pa->soa, 0,
 			     infobuff, sizeof(infobuff)));
 	 return -1;
       }
       Info1("permitting connection from %s due to range option",
-	    sockaddr_info((struct sockaddr *)pa, 0,
+	    sockaddr_info(&pa->soa, 0,
 			  infobuff, sizeof(infobuff)));
    }
 #endif /* WITH_IP4 */
@@ -1735,7 +1802,7 @@ int xiocheckpeer(xiosingle_t *xfd,
       if (pa->soa.sa_family == AF_INET &&
 	  ntohs(((struct sockaddr_in *)pa)->sin_port) != xfd->para.socket.ip.sourceport) {
 	 Warn1("refusing connection from %s due to wrong sourceport",
-	       sockaddr_info((struct sockaddr *)pa, 0,
+	       sockaddr_info(&pa->soa, 0,
 			     infobuff, sizeof(infobuff)));
 	 return -1;
       }
@@ -1744,20 +1811,20 @@ int xiocheckpeer(xiosingle_t *xfd,
       if (pa->soa.sa_family == AF_INET6 &&
 	  ntohs(((struct sockaddr_in6 *)pa)->sin6_port) != xfd->para.socket.ip.sourceport) {
 	 Warn1("refusing connection from %s due to wrong sourceport",
-	       sockaddr_info((struct sockaddr *)pa, 0,
+	       sockaddr_info(&pa->soa, 0,
 			     infobuff, sizeof(infobuff)));
 	 return -1;
       }
 #endif /* WITH_IP6 */
       Info1("permitting connection from %s due to sourceport option",
-	    sockaddr_info((struct sockaddr *)pa, 0,
+	    sockaddr_info(&pa->soa, 0,
 			  infobuff, sizeof(infobuff)));
    } else if (xfd->para.socket.ip.lowport) {
       if (pa == NULL)  { return -1; }
       if (pa->soa.sa_family == AF_INET &&
 	  ntohs(((struct sockaddr_in *)pa)->sin_port) >= IPPORT_RESERVED) {
 	 Warn1("refusing connection from %s due to lowport option",
-	       sockaddr_info((struct sockaddr *)pa, 0,
+	       sockaddr_info(&pa->soa, 0,
 			     infobuff, sizeof(infobuff)));
 	 return -1;
       }
@@ -1766,13 +1833,13 @@ int xiocheckpeer(xiosingle_t *xfd,
 	       ntohs(((struct sockaddr_in6 *)pa)->sin6_port) >=
 	       IPPORT_RESERVED) {
 	 Warn1("refusing connection from %s due to lowport option",
-	       sockaddr_info((struct sockaddr *)pa, 0,
+	       sockaddr_info(&pa->soa, 0,
 			     infobuff, sizeof(infobuff)));
 	 return -1;
       }
 #endif /* WITH_IP6 */
       Info1("permitting connection from %s due to lowport option",
-	    sockaddr_info((struct sockaddr *)pa, 0,
+	    sockaddr_info(&pa->soa, 0,
 			  infobuff, sizeof(infobuff)));
    }
 #endif /* WITH_TCP || WITH_UDP */
@@ -1782,12 +1849,12 @@ int xiocheckpeer(xiosingle_t *xfd,
    if (result < 0) {
       char infobuff[256];
       Warn1("refusing connection from %s due to tcpwrapper option",
-	    sockaddr_info((struct sockaddr *)pa, 0,
+	    sockaddr_info(&pa->soa, 0,
 			  infobuff, sizeof(infobuff)));
       return -1;
    } else if (result > 0) {
       Info1("permitting connection from %s due to tcpwrapper option",
-	    sockaddr_info((struct sockaddr *)pa, 0,
+	    sockaddr_info(&pa->soa, 0,
 			  infobuff, sizeof(infobuff)));
    }
 #endif /* (WITH_TCP || WITH_UDP) && WITH_LIBWRAP */
@@ -1954,9 +2021,9 @@ int xioparsenetwork(const char *rangename, int pf, struct xiorange *range) {
      strncpy(addrname, rangename, maskname-rangename-1);	/* ok */
      addrname[maskname-rangename-1] = '\0';
      result =
-	dalan(addrname, (char *)&range->netaddr.soa.sa_data, &addrlen,
+	dalan(addrname, (uint8_t *)&range->netaddr.soa.sa_data, &addrlen,
 	      sizeof(range->netaddr)-(size_t)(&((struct sockaddr *)0)->sa_data)
-	      /* data length */);
+	      /* data length */, 'i');
      if (result < 0) {
 	Error1("data too long: \"%s\"", addrname);
 	free(addrname); return STAT_NORETRY;
@@ -1966,9 +2033,9 @@ int xioparsenetwork(const char *rangename, int pf, struct xiorange *range) {
      }
      free(addrname); 
      result =
-	dalan(maskname, (char *)&range->netmask.soa.sa_data, &masklen,
+	dalan(maskname, (uint8_t *)&range->netmask.soa.sa_data, &masklen,
 	      sizeof(range->netaddr)-(size_t)(&((struct sockaddr *)0)->sa_data)
-	      /* data length */);
+	      /* data length */, 'i');
      if (result < 0) {
 	Error1("data too long: \"%s\"", maskname);
 	return STAT_NORETRY;
