@@ -1,5 +1,5 @@
 /* source: xioread.c */
-/* Copyright Gerhard Rieger 2001-2009 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* this is the source of the extended read function */
@@ -148,7 +148,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
        */
 #if defined(PF_PACKET) && defined(PACKET_OUTGOING)
       if (from.soa.sa_family == PF_PACKET) {
-	 if ((((struct sockaddr_ll *)&from.soa)->sll_pkttype & PACKET_OUTGOING)
+	 if ((from.ll.sll_pkttype & PACKET_OUTGOING)
 	    == 0) {
 	    errno = EAGAIN;  return -1;
 	 }
@@ -221,9 +221,12 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       }
 
       switch(from.soa.sa_family) {
+#if HAVE_STRUCT_IP
 	 int headlen;
+#endif /* HAVE_STRUCT_IP */
 #if WITH_IP4
       case AF_INET:
+#if HAVE_STRUCT_IP
 	 if (pipe->dtype & XIOREAD_RECV_SKIPIP) {
 	    /* IP4 raw sockets include the header when passing a packet to the
 	       application - we don't need it here. */
@@ -240,6 +243,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 	       bytes -= headlen;
 	    }
 	 }
+#endif /* HAVE_STRUCT_IP */
 	 break;
 #endif
 #if WITH_IP6
@@ -357,9 +361,12 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
       }
 
       switch(from.soa.sa_family) {
+#if HAVE_STRUCT_IP
 	 int headlen;
+#endif /* HAVE_STRUCT_IP */
 #if WITH_IP4
       case AF_INET:
+#if HAVE_STRUCT_IP
 	 if (pipe->dtype & XIOREAD_RECV_SKIPIP) {
 	    /* IP4 raw sockets include the header when passing a packet to the
 	       application - we don't need it here. */
@@ -376,6 +383,7 @@ ssize_t xioread(xiofile_t *file, void *buff, size_t bufsiz) {
 	       bytes -= headlen;
 	    }
 	 }
+#endif /* HAVE_STRUCT_IP */
 	 break;
 #endif
 #if WITH_IP6
