@@ -1,5 +1,5 @@
 /* source: xiosigchld.c */
-/* Copyright Gerhard Rieger */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* this is the source of the extended child signal handler */
@@ -11,6 +11,7 @@
 
 /*!! with socat, at most 4 exec children exist */
 pid_t diedunknown[NUMUNKNOWN];	/* children that died before they were registered */
+int   statunknown[NUMUNKNOWN]; 	/* exit state of unknown dead child */
 size_t nextunknown;
 
 
@@ -120,7 +121,8 @@ void childdied(int signum) {
 	 if (nextunknown == NUMUNKNOWN) {
 	    nextunknown = 0;
 	 }
-	 diedunknown[nextunknown++] = pid;
+	 diedunknown[nextunknown] = pid;
+	 statunknown[nextunknown++] = WEXITSTATUS(status);
 	 Debug1("saving pid in diedunknown"F_Zu,
 		nextunknown/*sic, for compatibility*/);
       }
@@ -130,12 +132,22 @@ void childdied(int signum) {
 	 Info2("waitpid(): child %d exited with status %d",
 	       pid, WEXITSTATUS(status));
       } else {
-	 Warn2("waitpid(): child %d exited with status %d",
-	       pid, WEXITSTATUS(status));
+	 if (i == XIO_MAXSOCK) {
+	    Info2("waitpid(): child %d exited with status %d",
+		   pid, WEXITSTATUS(status));
+	 } else {
+	    Error2("waitpid(): child %d exited with status %d",
+		   pid, WEXITSTATUS(status));
+	 }
       }
    } else if (WIFSIGNALED(status)) {
-      Info2("waitpid(): child %d exited on signal %d",
-	    pid, WTERMSIG(status));
+      if (i == XIO_MAXSOCK) {
+	 Info2("waitpid(): child %d exited on signal %d",
+	       pid, WTERMSIG(status));
+      } else {
+	 Error2("waitpid(): child %d exited on signal %d",
+	       pid, WTERMSIG(status));
+      }
    } else if (WIFSTOPPED(status)) {
       Info2("waitpid(): child %d stopped on signal %d",
 	    pid, WSTOPSIG(status));
